import os
import sys
import json
import logging
import asyncio

from app.app import Application

LOG_FORMAT = "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) -35s %(lineno) -5d: %(message)s"


def main():
    app_dir_name = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
    extra_config = {}
    if app_dir_name:
        extra_config = json.load(open("../{}.json".format(app_dir_name),
                                      encoding="utf8"))

    config = json.load(open("config/main.json", encoding="utf8"))
    error_conf = json.load(open("config/error_conf.json", encoding="utf8"))
    config = {**config, **error_conf, **extra_config}

    log_level = config["system"]["log_level"].upper()
    logging.basicConfig(
        level=log_level,
        format=LOG_FORMAT,
        filename=config["system"].get("log", None)
    )

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(log_level)
    ch.setFormatter(logging.Formatter(LOG_FORMAT))
    log = logging.getLogger()
    log.addHandler(ch)

    loop = asyncio.get_event_loop()
    app = Application(config=config, loop=loop)
    app.run()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import json
from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.schema.pipe_methods import (
    ext_http_in_tornado_request,
    pipe_head,
    pipe_check_skip,
    pipe_sep_string,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
)
from tracuni.misc.helper import (
    nerf_tail_digits
)


def get_response_body(response):
    response_body = getattr(response, 'response_body', None)
    if response_body is None:
        response_body = getattr(response, "_write_buffer", [b"{}"])
        response_body = json.loads(response_body[0].decode())
    return response_body


ruleset = (
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
            lambda path: path[0].get('full_path') if path else None,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, None),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Получить данные запроса и сохранить"
                    "для переиспользования",
        destination=Destination(DestinationSection.REUSE, 'request'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, ext_http_in_tornado_request),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Поставить флаг пропуска трассировки, если HTTP путь"
                    "в списке игнорируемых",
        destination=Destination(DestinationSection.REUSE, 'should_not_trace'),
        pipeline=(
            pipe_check_skip,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
            Origin(OriginSection.ADAPTER, None),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Разместить заголовки контекста трассера там, "
                    "где фабрика участков их ищет",
        destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
        pipeline=(
            pipe_head,
            lambda headers: dict((k.lower(), v) for k, v in (headers or {}).items()),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['zipkin_headers'],
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: [x[0], nerf_tail_digits(x[1])],
            lambda x: ("http", x[0].lower(), x[1]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['method'],
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Задать имя партнёрской точки для участка",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            lambda x: x[0] or x[1],
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['tracer_headers'].get('x-b3-servicename'),
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['remote_ip'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести адрес запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести хост из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.host',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_host', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести путь из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.path',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_path', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Сохранить информацию об ответе",
        destination=Destination(DestinationSection.REUSE, 'response'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, 'self'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги код статуса ответа",
        destination=Destination(DestinationSection.SPAN_TAGS, 'rsp.status'),
        pipeline=(
            pipe_head,
            lambda response: str(getattr(response, '_status_code', {})),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести параметры строки запроса в журнал участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_query_params'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            partial(pipe_head, **{'indexes': 1}),
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['query_str']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело запроса в журнал участка "
                    "- получить первый элемент источников"
                    "- выцепить существенные значения для тегов в строки"
                    "- сериализовать"
                    "- обрезать"
                    "- маскировать"
                    "- закавычить/откавычить/экранировать/энкодить",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_body'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            partial(pipe_head, **{'indexes': 1}),
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['body']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести в аннотации заголовки запроса",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.req.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['headers'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Разделитель запроса и ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'divider',
        ),
        pipeline=(
            lambda _: "--------------------------------",
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                None,
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести тело ответа на запрос",
        destination=Destination(DestinationSection.SPAN_LOGS, 'response_body'),
        pipeline=(
            pipe_head,
            get_response_body,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            partial(pipe_head, **{'indexes': 1}),
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в аннотации заголовки ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.rsp.headers',
        ),
        pipeline=(
            pipe_head,
            lambda response: getattr(response, '_headers', {}),
            lambda headers: dict(headers) if headers else None,
            pipe_dump,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести идентификатор виджета в теги",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'app.widget',
        ),
        pipeline=(
            pipe_head,
            lambda request: str(request.get('widget_id') or ''),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'request'),
        ),
        stage=Stage.PRE,
    ),
    # Rule(
    #     description="Запомнить ошибку для вывода",
    #     destination=Destination(DestinationSection.REUSE, 'error'),
    #     pipeline=(
    #         pipe_head,
    #     ),
    #     origins=(
    #         Origin(OriginSection.SPAN, 'error'),
    #     ),
    #     stage=Stage.POST,
    # ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Спецификация структур данных, используемых по всему коду библиотеки

Структура данных для извлечения
    * VariantToSchema Карта схем = Вариант - Схема
    * Variant Вариант = Направление + ВидИнтерфейса
    * Side Направление = К себе / От себя
    * APIKind ВидИнтерфейса = HTTP / ...(открытый список)
    * Schema Схема = список Правил
    * Rule Правило извлечения = Назначение + Источник + Фаза + Конвейер
    * Destination Назначение = РазделНазначения + имя метки
    * DestinationSection РазделНазначения = аннотации / метки заголовка ...
    * Origin Источник = РазделИсточника + функция / имя атрибута
    * OriginSection РазделИсточника = аргументы / результат / контейнер ...
    * Stage Фаза = Старт / До / После
    * Pipe Конвейер = список функций

"""

from enum import Enum
from functools import total_ordering

from typing import (
    NamedTuple,
    Iterable,
    Callable,
    Optional,
    Sequence,
    Dict,
    Union,
    Any,
    Tuple,
    Type,
)
from typing import TYPE_CHECKING

from .const import UNKNOWN_NAME

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from ..schema.engine.engine import (
        SchemaEngine,
    )  # pragma: no cover
    assert SchemaEngine  # pragma: no cover
    # noinspection PyUnresolvedReferences
    from ..adapter import TracerWrapper  # pragma: no cover
    # noinspection PyUnresolvedReferences
    from tracuni.point.span import SpanGeneral  # pragma: no cover


###############################################################################
# Вспомогательные типы


@total_ordering
class EnumAuto(Enum):
    """Добавить к старым версиям перечислений возможность автоматического
    присвоения индексных значений
    """

    def __new__(cls: Type['EnumAuto']) -> 'EnumAuto':
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __lt__(self, other: 'EnumAuto') -> Optional[bool]:
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


###############################################################################
# Структура конфигурации трассера и сервиса метрик и значения по умолчанию


StatsdOptionsMeasureNames = NamedTuple('StatsdOptionsMeasureNames', [
    ('check_count', Optional[str]),
    ('exception_count', Optional[str]),
    ('http_in_code_count', Optional[str]),
    ('http_in_method_time', Optional[str]),
    ('http_out_time', Optional[str]),
    ('db_time', Optional[str]),
])


StatsdOptionsMeasureNames.__new__.__defaults__ = (  # type: ignore
    'heartbeat', 'exception', 'gen_response_code', 'handler_timers', 'http_requests', 'db_query_function',
)


StatsdOptions = NamedTuple('StatsdOptions', [
    ('enable', bool),
    ('logging', bool),
    ('url', str),
    ('prefix', str),
    ('hb_interval', int),
    ('host', Optional[str]),
    ('port', Optional[int]),
    ('measure_names', Optional[StatsdOptionsMeasureNames]),
])
"""Спецификация настроек сервиса метрик

    Attributes
    ----------
    enable
        не писать в сервис метрик перехваченные от трассера данные
        и не инициализировать адаптер
    logging
        не писать в стандартный вывод диагностические сообщения от
        адаптера сервиса метрик
    url
        адрес сервиса метрик в формате "http://site.loc:1111"
    prefix
        добавлять ко всем ключам аннотации в начале данную строку
    hb_interval
        интервал проверки доступности сервиса (heartbeat)
    host
        часть строки url без номера порта
    port
        часть строки url, только номер порта
"""

StatsdOptions.__new__.__defaults__ = (  # type: ignore
    False, False, '', 'api', 2, None, None,
)
"""Настройки по умолчанию для сервиса метрик
"""

TracerOptions = NamedTuple('TracerOptions', [
    ('name', str),
    ('enable', bool),
    ('service_name', str),
    ('url', str),
    ('sample_rate', int),
    ('send_interval', int),
    ('skip_trace', Iterable[Any]),
    ('debug', bool),
    ('statsd', StatsdOptions),
    ('context_amqp_name', str),
])
"""Спецификация настроек сервиса метрик

    Attributes
    ----------
    name
        разновидность сервиса трассировки, пока всегда "zipkin"
    enable
        False - не писать события трассировки и не создавать адаптер,
        пропускать вызовы к обернутым функциям-точкам напрямую
    service_name
        название приложения в контексте трассера
    url
        адрес сервиса трассировки в формате "http://site.loc:1111/path/path/"
    sample_rate
        доля зарегистрированных событий, отправляемых трассеру
    send_interval
        как часто отправлять собранные данные трассеру
    skip_trace
        какие URL входящих точек HTTP не трассировать
    debug
        False - подавлять диагностические сообщения в стандартный вывод
    statsd
        настройки сервиса метрик хранятся внутри настроек трассера
"""

TracerOptions.__new__.__defaults__ = (  # type: ignore
    'zipkin', False, UNKNOWN_NAME, '', 1, 3, (), False, StatsdOptions(
        *StatsdOptions.__new__.__defaults__  # type: ignore
    ),
    None,
)
"""Настройки по умолчанию для трассера
"""


###############################################################################
# Варианты участков


# noinspection PyUnresolvedReferences
class SpanSide(EnumAuto):
    """Варианты направления точек

    Attributes
    ----------
    ALL
        любое направление, полезно для настройки правил извлечения содержимого
    IN
        входящая точка, запрос к нашему приложению со стороны
    OUT
        входящая точка, запрос от нашего приложения к стороннему
    """
    ALL = ()
    IN = ()
    OUT = ()
    MID = ()

# noinspection PyUnresolvedReferences
class APIKind(EnumAuto):
    """Варианты направления точек

    Attributes
    ----------
    ALL
        любое направление, полезно для настройки правил извлечения содержимого
    HTTP
        любые запросы по данному протоколу
    AMQP
        прием и отправка сообщений-задач через RabbitMQ очередь
    DB
        исходящие через адаптер базы данных запросы
    """
    ALL = ()
    HTTP = ()
    HTTPH = ()
    AMQP = ()
    RETASK = ()
    DB = ()
    INNER_RETRY = ()
    INNER_SHUTDOWN = ()
    LOG = ()
    CONSEQ = ()


class Part(EnumAuto):
    OPEN = ()
    CLOSE = ()
    CLOSE_AFTER = ()


Variant = NamedTuple('Variant', [
    ('span_side', SpanSide),
    ('api_kind', APIKind),
])
"""Задаёт определённые разновидности точек по направлению и виду интерфейса

    Применяется для:

    * задания конкретного вида обернутой точки
    * задания вида или видов точек, к которым применяется правило извлечения
    содержимого

    Attributes
    ----------
    span_side
        направление
    api_kind
        вид интерфейса
"""

Variant.__new__.__defaults__ = (SpanSide.ALL, APIKind.ALL)  # type: ignore
"""Настройки по умолчанию для варианта
"""

###############################################################################
# Структура настройки источника и назначения извлекаемых данных


# noinspection PyUnresolvedReferences
class OriginSection(EnumAuto):
    """Разделы источников извлекаемых данных

    Состоят из данных, имеющихся в оборачиваемой точке и из данных,
    создаваемых самой библиотекой. В действительности указывают на
    конкретные атрибуты класса управления точкой

    Attributes
    ----------
    POINT_ARGS
        действительные аргументы обернутой точки
        доступны, начиная с фазы INIT
    POINT_RESULT
        действительный возвращаемый результат обернутой точки
        доступны, начиная с фазы POST
    POINT_REF
        ссылка на саму обернутую точку (метод или функцию)
        доступны, начиная с фазы INIT
    CLIENT
        клиентский объект для доступа к экземпляру запроса и пр.
        доступны, начиная с фазы INIT
    ADAPTER
        обертка трассера из данной библиотеки
        доступны, начиная с фазы INIT
    PROCESSOR
        объект управления извлечением
        доступны, начиная с фазы INIT
    CALL_STACK
        сокращённый стек вызовов перед входом в точку
        доступны, начиная с фазы INIT
    REUSE
        словарь переиспользуемых значений
        доступны, начиная с фазы PRE
    SPAN
        ссылка на объект участка
        доступны, начиная с фазы PRE
    """
    POINT_ARGS = ()
    POINT_RESULT = ()
    POINT_REF = ()
    CLIENT = ()
    ADAPTER = ()
    ENGINE = ()
    CALL_STACK = ()
    REUSE = ()
    SPAN = ()
    LOGS = ()


Origin = NamedTuple('Origin', [
    ('section', OriginSection),
    ('getter', Union[str, Callable]),
])
"""Настройка источника извлечения данных

    Attributes
    ----------
    section
        секция - указатель на атрибут менеджера точки, где содержатся данные
    getter
        функция, принимающая на вход ссылку на секцию и возвращающая
        извлеченное значение, либо строка - наименование атрибута секции,
        содержащего данные
"""


# noinspection PyUnresolvedReferences
class DestinationSection(EnumAuto):
    """Разделы мест назначения извлекаемых данных

    Задают в какие части участка или в какие внутренние атрибуты попадут
    извлеченные данные

    Attributes
    ----------
    SPAN_TAGS
        Помещённые сюда строки, появятся в заголовке участка
        Размещает SpanGeneral
    SPAN_LOGS
        Помещённые сюда строки, появятся в аннотации участка
        Размещает SpanGeneral
    SPAN_NAME
        Помещённые сюда строки, появятся в названии участка
        Размещает SpanGeneral
    POINT_ARGS
        Запись в аргументы, передаваемые обёрнутой точке. !Внимание! Это
        единственное назначение, которое может непосредственно повлиять на
        работу клиентской программы
        Размещает PointAccessor
    ADAPTER
        Запись в атрибуты собственного адаптера. Забыл зачем потребовалось.
        Размещает PointAccessor
    REUSE
        Запись промежуточного хранения и повторного использования в других
        обработчиках уже извлечённых данных
        Размещает PointAccessor
    ALL
        Используется в командах, для указания, что она относится ко всем
        назначениям, не используется в конкретной схеме
        Мета-назначение, данные не размещаются

    """
    SPAN_TAGS = ()
    SPAN_LOGS = ()
    SPAN_NAME = ()
    POINT_ARGS = ()
    ADAPTER = ()
    REUSE = ()
    DEBUG = ()
    ALL = ()


Destination = NamedTuple('Destination', [
    ('section', DestinationSection),
    ('name', Optional[str]),
])
"""Настройка назначения извлечения данных

    Для настройки извлечения данный тип является ключом, то есть для каждого
    назначения используется только одна настройка

    Attributes
    ----------
    section
        секция - раздел места назначения извлекаемых данных
    name
        наименование атрибута, не для всех секций имеет значение
"""

Destination.__new__.__defaults__ = (None,)  # type: ignore
"""Настройки по умолчанию для назначения извлекаемых данных
"""

###############################################################################
# Информация по обернутой точке

PointArguments = NamedTuple('PointArguments', [

])


PointContext = NamedTuple('PointContext', [
    ("point_ref", Callable),
    ("adapter", 'TracerWrapper'),
    ("engine", 'SchemaEngine'),
    ("span", Optional['SpanGeneral']),
    ("point_args", Optional[Dict[str, Any]]),
    ("point_result", Optional[Any]),
    ("client", Optional[Any]),
    ("call_stack", Optional[Sequence]),
    ("reuse", Optional[Dict[str, Any]]),
    ("debug", Optional[Dict[str, Any]]),
    ("headers", Optional[Dict[str, Any]]),
    ("span_tags", Optional[Dict[str, str]]),
    ("span_logs", Optional[Sequence[str]]),
    ("span_name", Optional[Dict[str, str]]),
])
"""Контекст оборачиваемой точки вызова

    Структурой этого типа обмениваются аксессор точки и механизм извлечения,
    аксессор заполняет её исходными данными, механизм пишет в неё
    извлечённые, а также изменения в исходные данные. Применением этих
    данных к состоянию приложения и к отправляемым на трассер участкам
    занимается аксессор точки, формированием - механизм извлечения.

    Attributes
    ----------
        Все секции источников и назначения, кроме обощающих
"""

PointContext.__new__.__defaults__ = (None,) * 11  # type: ignore
"""Настройки по умолчанию для контекста точки, стек вызовов необязателен
"""


###############################################################################
# Структура данных фаз извлечения наполнения


# noinspection PyUnresolvedReferences
class Stage(EnumAuto):
    """Фазы извлечения содержимого участка

    На какой стадии извлекается содержимое (пост- или пре-) зависит от
    автоматического отнесения секции конкретной настройки извлечения к той или
    иной фазе или от явного указания фазы этой настройки

    Attributes
    ----------
    INIT
        перед вызовом до создания участка трассировки
    PRE
        перед вызовом после создания участка трассировки
    POST
        после вызова обернутой точки, доступны результаты её работы
    """
    INIT = ()
    PRE = ()
    POST = ()


MethodNameToStage = Dict[str, Stage]
"""Отнесение методов класса управления участком к фазам обработки наполнения
"""


###############################################################################
# Настройка обработки наполнения участков


# noinspection PyUnresolvedReferences
class PipeCommand(EnumAuto):
    """Значения командного элемента

    Использование элемента схемы как директивы управления

    Attributes
    ----------
    SKIP_LEVELS_BELOW
        Не использовать схемы того же варианта с более низким приоритетом
    SKIP_SECRET_MASK
        Не применять маскирование определённых полей дялзаданных нзначений
    """
    SKIP_LEVELS_BELOW = ()
    SKIP_SECRET_MASK = ()
    TEE = ()
    STOP = ()


TeeDescriptor = Dict[Union[OriginSection, DestinationSection], Any]
"""Структура упаковки дополнительных данных

    Указатель на то, куда их записать отображается на сами данные
"""

PipelineTee = NamedTuple('PipelineTee', [
    ('main_story', Any),
    ('side_story', TeeDescriptor),
])
"""Настройка дополнительной записи извлеченных данных посреди процесса
обработки

    Если функция конвейера вернет данные, обернутые в этот объект, то часть
    данных будет перезаписана в источнике или назначении. Что и чем будет
    перезаписано управляется содержимым side_story

    Этот тип предназначен для:
    * Отправить собранные значимые данные в метки участка
    * Добавить в HTTP заголовки запроса элементы трассера
    * Сохранить извлеченные данные для переиспользования другим правилом
    извлечения

Attributes
----------
main_story
    данные, идущие дальше по конвейеру
side_story
    куда и какие промежуточные данные записывать
"""

Origins = Tuple[Origin, ...]
"""Список источников данных
"""
Pipeline = Tuple[Union[Callable, PipeCommand, Destination], ...]
"""Список методов обработки извлечённых данных
"""

Rule = NamedTuple('SchemaRule', [
    ('destination', Destination),
    ('origins', Origins),
    ('pipeline', Pipeline),
    ('stage', Stage),
    ('description', Optional[str]),
])
"""Настройка извлечения, базовый элемент схемы

Тип элемента описания:
* в какой элемент содержимого участка их писать (секция и необязательное
наименование)
* откуда получать данные для заполнения (секция: SourceSection и путь:
строка или функция)
* как их при этом обрабатывать
* описание для справки
* необязательное явное указание этапа применения

Attributes
----------
description
    необязательное описание для документирования и ориентирования
stage
    явное указание фазы, в которой будет работать данная
    настройка
destination
    указывает секцию и наименование места, куда будут помещены извлеченные
    данные
origins
    указывает секцию и функция получения исходных данных
pipeline
    список методов, через обработку которыми проходит по цепочке данные
skip_pipe
    методы обработки, которые адо пропустить
"""

Rule.__new__.__defaults__ = ('', )  # type: ignore
"""Настройки по умолчанию для назначения извлекаемых данных
"""

RuleSet = Tuple[Rule, ...]
Extractors = Dict[Stage, RuleSet]
Schema = Dict[Variant, RuleSet]
"""Схема извлечения данных

    Отображение вариантов на списки настроек
    извлекаемых значений
"""



from tracuni.schema.builtin.core.ext_http_in import (
    ruleset as http_in_ruleset,
)
from tracuni.schema.builtin.core.ext_http_in_headers import (
    ruleset as http_in_headers_ruleset,
)
from tracuni.schema.builtin.core.ext_http_out import (
    ruleset as http_out_ruleset,
)
from tracuni.schema.builtin.core.ext_amqp_out import (
    ruleset as amqp_out_ruleset,
)
from tracuni.schema.builtin.core.ext_amqp_in import (
    ruleset as amqp_in_ruleset,
)
from tracuni.schema.builtin.core.ext_db_out import (
    ruleset as db_out_ruleset,
)
from tracuni.schema.builtin.core.ext_retry_out import (
    ruleset as retry_out_ruleset,
)
from tracuni.schema.builtin.tornado.ext_http_out import (
    ruleset as tornado_http_out_ruleset,
)
from tracuni.schema.builtin.tornado.ext_http_in import (
    ruleset as tornado_http_in_ruleset,
)
from tracuni.schema.builtin.tornado.ext_amqp_out import (
    ruleset as tornado_amqp_out_ruleset,
)
from tracuni.schema.builtin.tornado.ext_db_out import (
    ruleset as tornado_db_out_ruleset,
)

from tracuni.schema.builtin.tornado.ext_inner_retry_out import (
    ruleset as inner_retry_out_ruleset
)

from tracuni.schema.builtin.tornado.ext_inner_retry_in import (
    ruleset as inner_retry_in_ruleset
)

from tracuni.schema.builtin.tornado.log_out import (
    ruleset as log_out
)

__all__ = (
    'http_in_ruleset',
    'http_in_headers_ruleset',
    'http_out_ruleset',
    'amqp_in_ruleset',
    'amqp_out_ruleset',
    'db_out_ruleset',
    'retry_out_ruleset',
    'tornado_http_in_ruleset',
    'tornado_http_out_ruleset',
    'tornado_db_out_ruleset',
    'tornado_amqp_out_ruleset',
    'inner_retry_out_ruleset',
    'inner_retry_in_ruleset',
    'log_out',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Обработчик схем

    Управление схемами извлечения данных для различных вариантов точек

    Используется три уровня схем (в порядке убывания приоритета):
        * пользовательская на конкретную точку
            передаётся как инициализирующий параметр декоратора
        * пользовательская карта схем
            передаётся от пользователя как инициализирующий параметр
            адаптера библиотеки. Передаётся в процессор в момент его
            создания при инициализации декоратора
        * стандартная карта схем
            импортируется из schema.standard
"""

import inspect
import json

from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    Generator,
    MutableSequence,
    Callable,
)

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    PointContext,
    PipeCommand,
    APIKind,
)
from tracuni.misc.helper import json_dumps_decode

from tracuni.schema.engine.pool import SchemaEngineCache
from tracuni.schema.engine.feed import SchemaEngineFeed
from tracuni.misc.select_coroutine import get_coroutine_decorator
async_decorator = get_coroutine_decorator()

if TYPE_CHECKING:
    from typing import List  # noqa
    from typing import Union  # noqa
    from typing import Optional  # noqa
    from tracuni.define.type import Destination  # noqa
    from tracuni.define.type import MethodNameToStage  # noqa
    from tracuni.define.type import (  # noqa
        RuleSet,
        Extractors,
    )
    from tracuni.define.type import Stage
    from tracuni.adapter import AdapterType  # noqa
    from tracuni.point.accessor import PointAccessor


SUPPRESS_BD_SPANS = False


class SchemaEngine(metaclass=SchemaEngineCache):
    """Обработчик схемы. Связь варианта со схемой извлечения данных

        * Обращается к экземпляру поставщика схем для получения списка
        применяемых правил, разбитых на фазы и дополнительных данных
        * Для заданного экземпляра объекта доступа к точке извлекает и
        возвращает данные согласно правилам и состоянию объекта доступа.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.schema_feed = None  # type: SchemaEngineFeed
        self.adapter = None  # type: AdapterType
        self.extractors = None  # type: Extractors

        self.method_to_stage = None  # type: Optional[MethodNameToStage]
        self.skip_secret_destinations = (
            None  # type: Optional[MutableSequence[Destination]]
        )
        self.secret_mask_method_name = None  # type: Optional[str]

        self._args = args
        self._kwargs = kwargs

    def observe_state_change(self, adapter: 'AdapterType'):
        """
            Обработчик схемы создаётся до того, как отработает фабрика адаптера,
            соответственно вторая часть инициализации должна проходить после
            того, как адаптер будетсоздан

        Parameters
        ----------
        adapter
            только что созданный экземпляр адаптера
        """
        self.adapter = adapter
        self._kwargs['custom_schema'] = adapter.custom_schema
        self.schema_feed = schema_feed = SchemaEngineFeed(
            *self._args,
            **self._kwargs,
        )
        self.extractors = schema_feed.parse()  # type: Extractors

        self.skip_secret_destinations = (
            schema_feed.get_skip_secret()
        )  # type: MutableSequence[Destination]

        self.secret_mask_method_name = (
            schema_feed.get_secret_mask_method_name()
        )  # type: str

    def is_tracer_disabled(self):
        return (
            self.adapter is None
            or
            (
                SUPPRESS_BD_SPANS
                and
                self.schema_feed.variant.api_kind == APIKind.DB
            )
        )

    @async_decorator
    def extract(self,
                point: 'PointAccessor',
                stage: 'Stage',
                ) -> Generator[Any, None, PointContext]:
        """Метод запускающий процесс получения и изменения исходных данных и
        данных участка для указанной точки и этапа

        Parameters
        ----------
        point
            Ссылки на данные точки и окружения с ключами, совпадающими с
            наименованиями секций источников и назначения.
        stage
            Фаза, к которой привязан обратившийся метод

        Returns
        -------
            Изменённый контекст точки
        """
        point_context = point.context
        if stage is None:
            return point_context
        extractors = (
            self.extractors.get(stage) or []
        )  # type: Union[RuleSet, List]

        for extractor in extractors:
            extracted = yield from self._extract_values(extractor, point_context)
            if extracted is None or not len(extracted):
                continue
            point_context = self._pipe_values(extractor, extracted, point)

        return point_context

###############################################################################
# Area privata

    @staticmethod
    @async_decorator
    def _extract_values(
        extractor: Rule,
        point_context: PointContext
    ) -> Generator[Any, None, Tuple[Any, ...]]:
        """

        Parameters
        ----------
        extractor
            Применяемое в данный момент правило извлечения.
        point_context
            Структура, хранящая ссылки на контекст источник данных и уже
            собранные данные. Нужна для получения контекста источника.
        Returns
        -------
            Собранные исходные значения
        """
        val = []
        # правило может содержать несколько источников
        for origin in extractor.origins:
            extracted = None

            # получаем ссылку на источник или пропускаем источник
            section_ref = getattr(
                point_context, origin.section.name.lower(), None
            )
            if section_ref is None:
                continue

            # метод извлечения, либо функция, которой передается ссылка на
            # источник, а она возвращает значение по своему разумению
            getter = origin.getter
            if callable(getter):
                is_yields = inspect.isgeneratorfunction(getter)
                if is_yields:
                    extracted = yield from getter(section_ref)
                else:
                    extracted = getter(section_ref)
            # либо метод извлечения может быть указан строкой, тогда это или
            # элемент словаря, или атрибут объекта
            elif isinstance(getter, str):
                # отдельный случай прямой слэш - возращается исходный объект
                if getter == '/':
                    extracted = section_ref
                elif isinstance(section_ref, dict):
                    extracted = section_ref.get(getter)
                else:
                    extracted = getattr(section_ref, getter, None)
            elif getter is None:
                extracted = section_ref

            # if extracted is None:
            #     continue
            val.append(extracted)

        # данные, полученные из всех указанных в правиле источников,
        # собираются в кортеж
        return tuple(val)

    def _pipe_values(self, extractor: Rule, val: Any, point: 'PointAccessor') -> PointContext:
        """Конвейер обрботки извлекаемых значений

            Вызывает цепочку методов передавая от одного к другому данные в
            текущем состоянии, обрабатывает дополнительную команду ответвления
            данных и список подавленных обработчиков.

        Parameters
        ----------
        extractor
            Правило извелечения
        val
            Данные в текущем состоянии
        point
            Структура, хранящая ссылки на контекст источник данных и уже
            собранные данные.
        Returns
        -------
            Изменённый контекст точки
        """
        tee_val = None

        for idx, process in enumerate(extractor.pipeline):
            if self._skip_secret(process, extractor.destination):
                continue

            if isinstance(process, Destination):
                if val is None:
                    continue
                destination = process  # type: Destination
                output_value = val
                if destination.section == DestinationSection.DEBUG:
                    output_value = {
                        "value": json_dumps_decode(val, indent=2),
                        "processor": process,
                        "processor_idx": idx,
                        "extractor": extractor,
                    }
                point.update_context_by_destination(destination, output_value)
                continue

            if process == PipeCommand.TEE:
                tee_val = step_back_val
                continue

            step_back_val = val
            if tee_val is not None:
                val = tee_val
            val = process(val)
            tee_val = None

        point.update_context_by_destination(extractor.destination, val)
        return point.context

    def _skip_secret(self, fn: Callable, destination: "Destination") -> bool:
        """Проверка условия пропуска обработчика маскирования по команде

        Parameters
        ----------
        fn
            Обработчик
        destination
            Описание назначения для проверки применения команды

        Returns
        -------
            Флаг надо ли запускать данный обработчик
        """
        fn_name = getattr(fn, '__name__', '')
        if self.secret_mask_method_name != fn_name:
            return False

        return (
            (
                Destination(DestinationSection.ALL, "")
                in
                self.skip_secret_destinations
            )
            or
            destination in self.skip_secret_destinations
        )



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.misc.helper import nerf_tail_digits
from tracuni.schema.pipe_methods import (
    ext_http_in_aiohttp_request,
    pipe_head,
    pipe_check_skip,
    pipe_sep_string,
    pipe_catch_essential,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
    pipe_inject_headers,
    ext_out_headers,
)
from tracuni.define.errors import TracerCollectedException


def get_result_from_exception(x):
    if isinstance(x, TracerCollectedException):
        res = getattr(x, 'result', None)
    else:
        res = None
    return res or x


def decode_body(response):
    body = getattr(response, 'body', b'')
    try:
        body = body.decode()
    except AttributeError:
        try:
            body = body._value.decode()
        except AttributeError:
            pass
    return body


ruleset = (
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
            lambda path: path[0].get('full_path') if path else None,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, None),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Получить данные запроса и сохранить"
                    "для переиспользования",
        destination=Destination(DestinationSection.REUSE, 'request'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, ext_http_in_aiohttp_request),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Поставить флаг пропуска трассировки, если HTTP путь"
                    "в списке игнорируемых",
        destination=Destination(DestinationSection.REUSE, 'should_not_trace'),
        pipeline=(
            pipe_check_skip,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
            Origin(OriginSection.ADAPTER, None),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Разместить заголовки контекста трассера там, "
                    "где фабрика участков их ищет",
        destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
        pipeline=(
            pipe_head,
            lambda headers: dict((k.lower(), v) for k, v in (headers or {}).items()),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['zipkin_headers'],
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: [x[0], nerf_tail_digits(x[1]), x[2]],
            lambda x: (f"http::{x[2]}", x[0].lower(), x[1]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['method'],
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['mark'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Задать имя партнёрской точки для участка",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
        ),
        pipeline=(
            lambda x: x[0] or x[1],
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('tracer_headers', {}).get('x-b3-servicename', ''),
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['remote_ip'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести адрес запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести хост из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.host',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_host', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести путь из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.path',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_path', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Сохранить информацию об ответе",
        destination=Destination(DestinationSection.REUSE, 'response'),
        pipeline=(
            pipe_head,
            get_result_from_exception,
        ),
        origins=(
            Origin(OriginSection.POINT_RESULT, None),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги код статуса ответа",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'rsp.status'),
        pipeline=(
            pipe_head,
            pipe_head,
            lambda response: getattr(response, 'status', 500),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести параметры строки запроса в журнал участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_query_params'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            partial(pipe_head, **{'indexes': 1}),
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['query_str']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело запроса в аннотации участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_body'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['body']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести в логи заголовки запроса",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.req.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['headers'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Разделитель запроса и ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'divider',
        ),
        pipeline=(
            lambda _: "--------------------------------",
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                None,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело ответа на запрос",
        destination=Destination(DestinationSection.SPAN_LOGS, 'response_body'),
        pipeline=(
            pipe_head,
            pipe_head,
            # lambda response: getattr(response, 'body', b'').decode(),
            decode_body,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в логи заголовки запроса",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.req.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['headers'],
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в логи заголовки ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.rsp.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_head,
            lambda response: getattr(response, 'headers', {}),
            lambda headers: dict(headers) if headers else None,
            pipe_dump,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести идентификатор виджета в теги",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'app.widget',
        ),
        pipeline=(
            pipe_head,
            lambda request: str(request.get('widget_id') or ''),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'request'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Переиспользование экземпляров механизма извлечения

    Для каждого используемого варианта экземпляр создаётся один раз.
    Исключение составляет случай, когда передаётся набор правил точки,
    для упрощения, чтобы не надо было проверять одинаковые это
    пользовательские правила для точки или нет.
"""

from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
)

from ...define.type import (
    Variant,
    RuleSet,
    Rule,
    Schema,
)
from ...define import errors as err

if TYPE_CHECKING:
    from typing import Dict  # noqa
    from typing import Optional  # noqa


class SchemaEngineCache(type):
    """Сохранение экземпляров мастера схем для переиспользования

        Не сохраняет, если задана пользовательская схема
    """

    def __init__(cls, *args: Any, **kwargs: Any):
        """Перекрывается для создания атрибута хранения кеша
        """
        super().__init__(*args, **kwargs)
        cls._cache = {}  # type: Dict[Variant, Any]

    def __call__(
        cls: 'SchemaEngineCache',
        *args: Any,
        **kwargs: Any
    ) -> 'SchemaEngineCache':
        """Перекрывается для работы с кешем при создании процессора

            * получает инициализирующие параметры и проверяет их
            * берет экземпляр процессора из кеша, если нет пользовательской
            схемы, а он там есть
            * создаёт экземпляр
            * записывает  созданный экземпляр в кеш, если нет
            пользовательской схемы

        Parameters
        ----------
        variant (или нулевой в позиционных аргументах)
            используется как ключ кеша для созданного экземпляра
        custom_schema (или первый в позиционных аргументах)
            проверяется наличие, чтобы не создавать кеш для процессора с
            пользовательской схемой

        Returns
        -------
            Созданный экземпляр процессора
        """
        self = None  # type: Optional[SchemaEngineCache]

        variant, point_ruleset, _ = cls._provide_init_values(
            args, kwargs
        )

        if point_ruleset is None:
            self = cls._cache.get(variant)
        if self is None:
            self = super().__call__(*args, **kwargs)
            if point_ruleset is None:
                cls._cache[variant] = self

        return self

    def _provide_init_values(
        self,
        args: Any, kwargs: Any
    ) -> Tuple[Variant, RuleSet, Schema]:
        """Обработка параметров инициализации конкретного объекта

            Получает параметры варианта или пользовательской схемы из
            позиционных или именованных аргументов экземпляра дочернего
            объекта.

        Parameters
        ----------
        variant (или нулевой в позиционных аргументах)
            используется как ключ кеша для созданного экземпляра
        custom_schema (или первый в позиционных аргументах)
            проверяется наличие, чтобы не создавать кеш для процессора с
            пользовательской схемой
        custom_schema_map (или второй в позиционных аргументах)
            на уровне мета-класса только проверяется

        Returns
        -------
            Возвращает кортеж из полученных параметров
        """
        variant = (args[0:1] or [None])[0]  # type: Optional[Variant]
        if variant is None:
            variant = kwargs.get('variant')
        point_ruleset = (args[1:2] or [None])[0]  # type: Optional[RuleSet]
        if point_ruleset is None:
            point_ruleset = kwargs.get('point_ruleset')
        custom_schema = (
            args[2:3] or [None]
        )[0]  # type: Optional[Schema]
        if custom_schema is None:
            custom_schema = kwargs.get('custom_schema')

        return self._check_arguments(variant, point_ruleset, custom_schema)

    @staticmethod
    def _check_arguments(
        variant: Variant,
        point_ruleset: RuleSet,
        custom_schema: Schema,
    ) -> Tuple[Variant, RuleSet, Schema]:
        """Проверяет инициализирующие аргументы

        Parameters
        ----------
        variant
            Проверяемый аргумент, должен присутствовать
        point_ruleset
        custom_schema
            Проверяемый аргумент, необязателен

        Returns
        -------
            Те же аргументы, что и на входе

        Raises
        -------
            Ошибку проверки наличия или типа того или иного аргумента
        """
        if not isinstance(variant, Variant):
            raise err.ProcessorInitNoneOrWrongTypeVariant

        if (
            point_ruleset
            and
            (
                not isinstance(point_ruleset, tuple)
                or
                any(type(el) != Rule for el in point_ruleset)
            )
        ):
            raise err.ProcessorInitWrongTypeCustomSchema

        if (
            custom_schema
            and
            (
                not isinstance(custom_schema, dict)
                or
                any(type(el) != Variant for el in custom_schema.keys())
                or
                any(
                    type(schema_item) != Rule
                    for schema in custom_schema.values()
                    for schema_item in schema
                )
            )
        ):
            raise err.ProcessorInitWrongTypeCustomSchema

        return variant, point_ruleset, custom_schema



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Тесты модуля доступа к вызываемой точке

"""

import pytest


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


###############################################################################
# Настройки трассера и сервиса сбора метрик


@pytest.fixture
def point_accessor_type():
    """
    """
    # mocker.patch('trac_uni.const.UNKNOWN_NAME', unknown_name)
    from tracuni.trac_uni import PointAccessor
    return PointAccessor


def test_class_exist(point_accessor_type):
    assert hasattr(point_accessor_type({}), 'point_context')



import asyncio
import json
import logging
from abc import ABCMeta

import aioamqp

from app.core.helper import json_encode


class BaseHandler(object):
    __metaclass__ = ABCMeta

    def __init__(self, server):
        self.server = server
        self.loop = None

    @property
    def app(self):
        return self.server.app

    @property
    def config(self):
        return self.server.app.config

    async def task(self, name, task):
        time_waited = 0
        to_wait = 1.5
        while True:
            try:
                transport, protocol = await aioamqp.connect(
                    **self.config['rabbit']
                )
                break
            except (OSError, aioamqp.AmqpClosedConnection) as e:
                to_wait = round(min(30, (to_wait ** 1.5)), 2)
                logging.error(
                    "[x] Failed to connect to tasks RabbitMQ: %s. Waiting %s seconds...", e, to_wait)
                await asyncio.sleep(to_wait)
                time_waited += to_wait
        channel = await protocol.channel()

        sub_tasks = task.get('tasks', {})
        data = {
            "name": name,
            "tasks": sub_tasks,
        }
        res = await channel.publish(
            json_encode(data),
            task.get('address'),
            self.config['rabbit']['routing_key']
        )

        await protocol.close()
        # ensure the socket is closed.
        transport.close()
        logging.info('[i] Task sended')

        return {
            "name": name,
            "ok": True,
            "tasks": sub_tasks,
            "result": {
                "message": "task sent",
            },
        }



import asyncio
import json

from aiohttp import web
from aiohttp.web_request import Request
from uuid import uuid4

from app.core.http_server import BaseHandler
from app.core.helper import json_encode
from app.lib.helper import Helper


class HttpHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.http_cln = self.app.http_cln
        self.helper = Helper(self.app.config, None)

    async def http_in(self, context_span, request: Request):
        req = await self.helper.check_request(request, 'transfer')
        waiting_for = []
        for name, task in req.get('tasks', {}).items():
            sub_tasks = task.get('tasks', {})
            if task.get('api_kind') == 'http':
                url = task.get('address')
                if url:
                    http_res = self.app.http_cln.request(
                        "POST",
                        url,
                        headers={},
                        body=json.dumps({
                            "name": name,
                            "tasks": sub_tasks
                        }).encode(),
                    )
                    waiting_for.append(http_res)
            if task.get('api_kind') == 'amqp':
                amqp_res = self.task(name, task)
                waiting_for.append(amqp_res)
            if task.get('api_kind') == 'db':
                db_res = getattr(self.app.db_tasks, task.get('address'))(
                    name, task)
                waiting_for.append(db_res)

        if waiting_for:
            response = await asyncio.gather(*waiting_for)
        else:
            response = []
        response = [el for el in response if el is not None]
        response = {
            "name": req.get("name", "unknown"),
            "ok": True,
            "cvv": 123,
            "tasks": dict(
                (
                    response_item.get("name", uuid4().hex),
                    dict(
                        (k, v)
                        for k, v in response_item.items()
                        if k != "name"
                    )
                )
                for response_item in response
            ),
        }

        return web.json_response(response, dumps=json_encode)

    async def callback_in(self, context_span, request: Request):
        req = await self.helper.check_request(request, 'transfer')
        self.app.log_info("Result: {}".format(
            json_encode(req)
        ))

    async def gen_error(self, _, request: Request):
        raise Exception("1")




from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.pipe_methods import (
    pipe_head,
    pipe_sep_string,
    ext_tracer_path,
    ext_tracer_short_path,
    ext_tracer_full_path
)
from tracuni.schema.shared_rules import rule_inner_out_tracer_context, rule_span_peer_name_from_application_name

ruleset = (
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.REUSE, 'point_name'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_short_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Путь по методам, вызвавшим запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_full_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: ("app", "retry", x[0]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, 'point_name',
            ),
        ),
        stage=Stage.PRE,
    ),
    rule_inner_out_tracer_context,
    rule_span_peer_name_from_application_name,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Обработка участков на каждом этапе
"""
import sys
from functools import wraps

from tracuni.define.type import (
    SpanSide,
    Stage,
)
import tracuni.define.errors as err
from tracuni.misc.select_coroutine import get_coroutine_decorator
from tracuni.misc.helper import log_tracer_error

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from aiozipkin.span import Span
    from typing import Dict
    from tracuni.define.type import MethodNameToStage
async_decorator = get_coroutine_decorator()


class ProvideSpanForStage:
    """Декоратор для методов этапа
            * Гарантирует наличие экземпляра участка
            * Получает наполнение для текущей стадии
            * Сохранят результат вызова точки

        Parameters
        ----------
        stage
            Привязывает декорированный метод к конкретному этапу. Если таких
            несколько, то привязывается только первый обработанный

        Returns
        -------
            Метод обработки определённого этапа
    """
    # Отношение наименований методов аксессора точки к запускаемым фазам
    # извлечения, отрабатывающим до выполнения этих методов.
    method_to_stage = {
        # 'run_init_stage': Stage.INIT,
        # 'run_pre_stage': Stage.PRE,
        # 'run_post_stage': Stage.POST,
    }  # type: MethodNameToStage

    stage_to_method = {}  # type: Dict[Stage, str]

    def __init__(self, stage: Stage):
        self.stage = stage
        self.span_general = None

    def __call__(self, fn):
        # Сохраняем прямое и обраное отображение этапов на методы обработки
        self.fn_name = fn_name = fn.__name__
        if fn_name and self.stage not in self.method_to_stage:
            self.method_to_stage[fn_name] = self.stage
            self.stage_to_method[self.stage] = fn_name

        @wraps(fn)
        @async_decorator
        def wrapper(span_general):
            try:
                self.span_general = span_general
                self.point = span_general.point

                # Получаем текущий этап
                stage_found = self._detect_stage()
                # Запускаем дополнительные операции для данного этапа
                self._stage_operations(stage_found)
                # Запускаем основной конвейер извлечения данных для участка
                yield from self._run_stage_pipe(stage_found)
                # Пропускаем данный участок и его детей, если это входящий HTTP, чей адрес указан в опции skip
                self._skip_http()
                # Создаём экземпляр участка, если его ещё нет
                span_instance = self._create_span_inst_if_none()

                # Запускаем метод формирования участка на данном этапе
                yield fn(span_general, span_instance)

            except Exception as exc:
                debug_is_on = self.point.context.adapter.config.debug
                log_tracer_error(exc, debug_is_on)
                if self.point:
                    self.point.register_error(exc, sys.exc_info())
                span_general.emergency_exit(exc)

            return

        return wrapper

###############################################################################
# Area privata
    def provide_span_and_point(fn):
        def wrapper(self, *args, **kwargs):
            span_general = self.span_general
            point = getattr(span_general, 'point', None)
            if not point:
                raise err.NoSpanException
            return fn(self, span_general, point, *args, **kwargs)
        return wrapper

    @provide_span_and_point
    def _stage_operations(self, span_general, point, stage_found):
        _ = span_general

        # Для этапа после вызова точки сохраняем результат её работы
        if stage_found == Stage.POST:
            result = None
            if (
                not result
                and
                hasattr(
                    point.context.point_args,
                    'self',
                )
                and
                hasattr(
                    point.context.point_args.self,
                    'response_body',
                )
            ):
                result = point.context.point_args.self.response_body
            if point.context.point_result is None:
                point.update_context({"point_result": result})

    @provide_span_and_point
    def _detect_stage(self, span_general, point):
        _ = span_general
        _ = point

        # Определяем текущий этап
        stage_found = self.method_to_stage.get(self.fn_name)
        if not stage_found:
            raise err.SpanNoStageForMethod

        return stage_found

    @provide_span_and_point
    def _skip_http(self, span_general, point):
        _ = span_general
        # обрабатываем выставленный обработчиками флаг пропуска
        if point.context.reuse.get('should_not_trace'):
            raise err.HTTPPathSkip

    @provide_span_and_point
    @async_decorator
    def _run_stage_pipe(self, span_general, point, stage_found):
        _ = span_general
        # Вызываем извлечение и обработку данных этапа
        pipe_fn = point.context.engine.extract

        point_context = yield from pipe_fn(point, stage_found)
        if point_context is None:
            point_context = point.context
        point.context = point_context

    @provide_span_and_point
    def _create_span_inst_if_none(self, span_general, point):
        # создаём участок, если он ещё не создан
        span_instance: Span = getattr(
            span_general,
            'span_instance',
            None
        )
        if span_instance is None:
            if span_general.side == SpanSide.IN:
                span_instance = span_general.fab_in_span()
            elif span_general.side in (SpanSide.OUT, SpanSide.MID,):
                span_instance = span_general.fab_out_span()

            if span_instance is None:
                raise err.NoSpanException
            span_general.span_instance = span_instance

        if not point.context.span:
            point.update_context({"span": span_general})

        return span_instance



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.misc.helper import nerf_tail_digits
from tracuni.schema.pipe_methods import (
    ext_out_headers,
    ext_tracer_out_point,
    ext_http_out_url,
    pipe_inject_headers,
    pipe_head,
    pipe_sep_string,
    pipe_service_name_from_config,
    pipe_catch_essential,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
    ext_send_request_args,
    ext_api_kind,
    pipe_get_send_request_core,
    ext_tracer_path,
    ext_tracer_full_path,
)


def deb(x):
    return x


ruleset = (
    Rule(
        description="Путь по методам, вызвавшим запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_full_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Сохранить данные запроса из аргументов для переиспользвоания",
        destination=Destination(DestinationSection.REUSE, 'request'),
        pipeline=(
            pipe_head,
            pipe_get_send_request_core,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, None),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: [x[0], nerf_tail_digits(x[1])],
            lambda x: ("http", x[0].lower(), x[1]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['method'],
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Записать заголовки в запрос",
        destination=Destination(DestinationSection.POINT_ARGS, 'headers'),
        pipeline=(
            pipe_inject_headers,
        ),
        origins=(
            Origin(
                OriginSection.SPAN, None,
            ),
            Origin(
                OriginSection.POINT_ARGS, ext_out_headers,
            ),
            Origin(
                OriginSection.ADAPTER,
                lambda adapter: adapter.config.service_name,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Сохранить информацию об ответе",
        destination=Destination(DestinationSection.REUSE, 'response'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_RESULT, None),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги код статуса ответа",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'rsp.status'),
        pipeline=(
            pipe_head,
            lambda response: (response or {}).get('http_code', 666),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести параметры строки запроса в журнал участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_query_params'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            partial(pipe_head, **{'indexes': 1}),
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['query']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело запроса в аннотации участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_body'),
        pipeline=(
            pipe_head,
            lambda x: x or '',
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['body']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести в логи заголовки запроса",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.req.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['headers'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Разделитель запроса и ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'divider',
        ),
        pipeline=(
            lambda _: "--------------------------------",
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                None,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело ответа на запрос",
        destination=Destination(DestinationSection.SPAN_LOGS, 'response_body'),
        pipeline=(
            pipe_head,
            deb,
            lambda response: (response or {}).get('response', response if type(response) == dict else ''),
            deb,
            pipe_dump,
            deb,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в логи заголовки ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'http.rsp.headers',
        ),
        pipeline=(
            pipe_head,
            lambda response: (response or {}).get('response_headers', {}),
            lambda headers: dict(headers) if headers else None,
            pipe_dump,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Наименование партнерского сервиса сохранить для "
                    "переиспользования",
        destination=Destination(
            DestinationSection.REUSE,
            'peer_name',
          ),
        pipeline=(
            pipe_service_name_from_config,
        ),
        origins=(
            Origin(OriginSection.ADAPTER, 'url_service_map'),
            Origin(OriginSection.POINT_ARGS, ext_http_out_url),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Наименование партнерского сервиса записать в участок",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'peer_name'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Хост партнерского сервиса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'peer.url',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, ext_http_out_url),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести адрес запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести хост из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.hostname',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_hostname', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести адрес из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.address',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_address', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести путь из адреса запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'rq.url.path',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('request', {}).get('url_path', ''),
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Привязка наборов правил к вариантам

    variant => ruleset

    ruleset = (
        Rule(
            description='',
            destination=Destination(
                section=DestinationSection.SPAN_TAGS,
                name='some'
            ),
            stage=Stage.PRE,
            origins=(
                Origin(section=OriginSection.REUSE, getter='some'),
            ),
            pipeline=(lambda x: x,),
        ),
    )  # type: RuleSet
"""
from typing import TYPE_CHECKING

from tracuni.misc.select_coroutine import is_tornado_on
from tracuni.define.type import (
    Variant,
    SpanSide,
    APIKind,
    Rule,
    Destination,
    DestinationSection,
    Stage,
    Origin,
    OriginSection,
)
from tracuni.schema.pipe_methods import pipe_inject_headers

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from ..define.type import RuleSet  # noqa
    from ..define.type import Schema  # noqa


####################
# Подключение стандартных настроек экстракторов

# для варианта будет использован только один из наборов для определённого
# ключа назначения в зависимости от разновидности ключа экстракторы
# применяются в следующем приоритете сверху вниз (если применяется верхний,
# то нижний уже не применяется)
#  - элементы с конкретным вариантом применяются
# только для этого варианта
#  - элементы привязанные только к направлению применяются ко всем вариантам
#  с этим направлением
#  - элементы привязанные по виду API применяются ко всем вариантам с этим
#  видом
#  - непривязанные к вариантам элементы применяются ко всем вариантам

conseq_out_ruleset = (
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda _: "CONSEQUENTIAL BLOCK",
        ),
        origins=(
            Origin(
                OriginSection.SPAN, None,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Передать заголовки в обработчик",
        destination=Destination(DestinationSection.POINT_ARGS, 'headers'),
        pipeline=(
            pipe_inject_headers,
        ),
        origins=(
            Origin(
                OriginSection.SPAN, None,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Задать имя партнёрской точки для участка",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
        ),
        pipeline=(
            lambda _: "many peers",
        ),
        origins=(
            Origin(
                OriginSection.SPAN, None,
            ),
        ),
        stage=Stage.PRE,
    ),
)


def get_standard_schema():
    from tracuni.schema.builtin import (
        inner_retry_out_ruleset,
        inner_retry_in_ruleset,
        log_out,
    )
    if is_tornado_on():
        from tracuni.schema.builtin import (
            tornado_http_in_ruleset as http_in_ruleset,
            tornado_http_in_ruleset as http_in_headers_ruleset,
            tornado_http_out_ruleset as http_out_ruleset,
            tornado_db_out_ruleset as db_out_ruleset,
            tornado_amqp_out_ruleset as amqp_out_ruleset,
            amqp_in_ruleset,
            retry_out_ruleset,
        )
    else:
        from tracuni.schema.builtin import (
            http_in_ruleset,
            http_in_headers_ruleset,
            http_out_ruleset,
            amqp_in_ruleset,
            amqp_out_ruleset,
            db_out_ruleset,
            retry_out_ruleset,
        )
    standard_schema = {
        Variant(SpanSide.IN, APIKind.HTTP): http_in_ruleset,
        Variant(SpanSide.IN, APIKind.HTTPH): http_in_headers_ruleset,
        Variant(SpanSide.OUT, APIKind.HTTP): http_out_ruleset,
        Variant(SpanSide.IN, APIKind.AMQP): amqp_in_ruleset,
        Variant(SpanSide.OUT, APIKind.AMQP): amqp_out_ruleset,
        Variant(SpanSide.OUT, APIKind.DB): db_out_ruleset,
        Variant(SpanSide.OUT, APIKind.RETASK): retry_out_ruleset,
        Variant(SpanSide.IN, APIKind.INNER_RETRY): inner_retry_in_ruleset,
        Variant(SpanSide.OUT, APIKind.INNER_RETRY): inner_retry_out_ruleset,
        Variant(SpanSide.MID, APIKind.LOG): log_out,
        Variant(SpanSide.IN, APIKind.INNER_SHUTDOWN): log_out,
        Variant(SpanSide.OUT, APIKind.CONSEQ): conseq_out_ruleset,
    }  # type: Schema
    return standard_schema





import hashlib
import hmac
import json
from urllib.parse import parse_qs
from aiohttp.web_request import Request
from app.core.error import Error
from app.core.http_server import SPAN_KEY


class CheckError(Error):
    def __init__(self, code):
        self.code = code


class Helper:
    def __init__(self, config, db_transfer):
        self.config = config
        self.db_transfer = db_transfer

    def get_status_message(self, code):
        return self.config["status_messages"][str(code)]

    def get_response_message(self, code):
        return self.config["errors"][str(code)]

    def gen_response(self, code):
        return {"code": code, "message": self.get_response_message(code)}

    def gen_status(self, code):
        return {"code": code, "message": self.get_status_message(code)}

    @staticmethod
    def check_req(req, method):
        res = False
        if method == "transfer":
            if "from_msisdn" in req and "to_msisdn" in req and "amount" in req:
                res = True
        elif method == "transfer_card":
            if "from_msisdn" in req and "to_card" in req and "amount" in req:
                res = True
        elif method == "transfer_account":
            if "from_msisdn" in req and "account" in req and "amount" in req:
                res = True
        elif method == "transfer_custom":
            if "from_msisdn" in req and "gateway_id" in req and "amount" in req and "fields" in req:
                res = True
        elif method == "status":
            if "transfer_id" in req or "merc_pid" in req:
                res = True
        elif method == "history":
            if "client_id" in req and "from_stamp" in req and "till_stamp" in req:
                res = True
        elif method == "calculate":
            if "client_id" in req and "amount" in req and "target" in req:
                res = True
        elif method == "get_gateway":
            res = True
        elif method == 'get_suggestions':
            if "query" not in req:
                raise CheckError(24)
            if "kind" not in req:
                raise CheckError(25)
            res = True
        elif method == 'create_template':
            if "client_id" not in req:
                raise CheckError(28)
            if "name" not in req:
                raise CheckError(26)
            if "oper_order_id" not in req:
                raise CheckError(27)
            res = True
        elif method == 'get_templates':
            if "client_id" not in req:
                raise CheckError(28)
            res = True
        elif method == 'delete_template':
            if "client_id" not in req:
                raise CheckError(28)
            if "template_uuid" not in req:
                raise CheckError(29)
            res = True
        return res

    async def check_request(self, request: Request, method: str):
        body = await request.text()

        if request.method == 'POST':
            try:
                req = json.loads(body)
            except:
                raise CheckError(4)
        else:
            req = {}
            _req = parse_qs(request.query_string)
            for k, v in _req.items():
                req[k] = v[0]

        return req

    async def authenticate(self, request: Request):
        # Аутентификация мерчанта
        context_span = request[SPAN_KEY]
        body = await request.text()
        params = request.url.query
        if "api_key" in params and "sign" in params:
            # проверка подписи
            secret_key = await self.db_transfer.get_secret(context_span, params["api_key"])
            if secret_key:
                check_sign = hmac.new(secret_key.encode(), body.encode(), hashlib.sha256).hexdigest()
                if params["sign"] == check_sign:
                    merc_data = await self.db_transfer.get_mercdata_by_apikey(context_span, params["api_key"])
                    response = self.gen_response(0)
                else:
                    return self.gen_response(3)
            else:
                return self.gen_response(3)
        else:
            ssl_verified = request.headers.get("Ssl_Verified", None)
            ssl_serial = int(request.headers.get("Ssl_Serial", "0"), 16)
            ssl_idn = request.headers.get("Ssl_Idn", None)

            if ssl_verified != "SUCCESS":
                return self.gen_response(5)

            check_serial = await self.db_transfer.check_serial(context_span, ssl_serial)
            if check_serial:
                merc_data = await self.db_transfer.get_mercdata_by_cert(context_span, ssl_serial)
                response = self.gen_response(0)
            else:
                return self.gen_response(2)

        response = {**response, **merc_data}
        return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# noinspection PyProtectedMember

"""Проверка кеширования обработчиков схем
"""

import pytest

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from tracuni.trac_uni import (
        SchemaEngineCache,
    )  # noqa


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_proc_cache_init(
    engine_types,
    variants,
):
    e_c_t = engine_types[1]  # type: SchemaEngineCache
    err = engine_types[3]
    test_variant, test_another_variant, *_ = variants

    class Processor(metaclass=e_c_t):
        def __init__(self, *args, **kwargs):
            pass

    with pytest.raises(
        err.ProcessorInitNoneOrWrongTypeVariant,
    ) as exc_info:
        invalid_processor = Processor()
        del invalid_processor
    assert (
        exc_info
    ), "Нельзя создать процессор без варианта"

    with pytest.raises(
        err.ProcessorInitNoneOrWrongTypeVariant,
    ) as exc_info:
        invalid_processor = Processor('wrong')
        del invalid_processor
    assert (
        exc_info
    ), "Нельзя создать процессор с неправильным вариантом"

    processor = Processor(
        test_variant,
    )
    assert (
        processor
    ), "Можно создать процессор без пользовательских схем"

    processor_cache = getattr(processor, '_cache', None)
    assert (
        isinstance(processor_cache, dict)
    ), "Содержит словарный атрибут кеша"
    assert (
        processor_cache
    ), "Кешируется, если нет пользовательской схемы"

    processor_the_same_variant = Processor(
        test_variant,
    )
    assert (
        processor_the_same_variant._cache is processor._cache
    ), "Кеш един для обоих экземпляров"
    assert (
        len(processor_cache) == 1
    ), "Не создаётся новая запись в кеше для того же варианта"

    Processor(
        test_another_variant,
    )
    assert (
        len(processor_cache) == 2
    ), "Создаётся новая запись в кеше для другого варианта"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_proc_cache_init_cust_point(
    engine_types,
    point_ruleset,
    variants,
):
    (
        _,
        e_c_t,  # type: SchemaEngineCache
        _,
        err,
    ) = engine_types
    ruleset, invalid_ruleset, *_ = point_ruleset
    test_variant, *_ = variants

    class Processor(metaclass=e_c_t):
        def __init__(self, *args, **kwargs):
            pass

    processor = Processor(
        test_variant,
        ruleset,
    )

    processor_cache = getattr(processor, '_cache', None)
    assert (
        isinstance(processor_cache, dict)
    ), "Содержит словарный атрибут кеша"
    assert (
        not processor_cache
    ), "Не кешируется, если инициализируется с пользовательской схемой точки"

    with pytest.raises(
        err.ProcessorInitWrongTypeCustomSchema,
    ) as exc_info:
        invalid_processor = Processor(
            test_variant,
            'wrong'
        )
        del invalid_processor
    assert (
        exc_info
    ), "Отлавливается ошибка, когда схема полностью некорректна"

    with pytest.raises(
        err.ProcessorInitWrongTypeCustomSchema,
    ) as exc_info:
        invalid_processor = Processor(
            test_variant,
            invalid_ruleset,
        )
        del invalid_processor
    assert (
        exc_info
    ), "Отлавливается ошибка, когда один из элементов схемы некорректен"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_proc_cache_init_cust_map(
    engine_types,
    point_ruleset,
    variants,
):
    _, e_c_t, _, err = engine_types
    ruleset, invalid_ruleset, *_ = point_ruleset
    test_variant, *_ = variants

    class Processor(metaclass=e_c_t):
        def __init__(self, *args, **kwargs):
            pass

    processor = Processor(
        test_variant,
        custom_schema={
            test_variant: ruleset
        }
    )

    processor_cache = getattr(processor, '_cache', None)
    assert (
        isinstance(processor_cache, dict)
    ), "Содержит словарный атрибут кеша"
    assert (
        processor_cache
    ), "Кешируется, если инициализируется с пользовательской картой схем"

    with pytest.raises(
        err.ProcessorInitWrongTypeCustomSchema,
    ) as exc_info:
        invalid_processor = Processor(
            test_variant,
            custom_schema='wrong'
        )
        del invalid_processor
    assert (
        exc_info
    ), "Отлавливается ошибка, когда карта полностью некорректна"

    with pytest.raises(
        err.ProcessorInitWrongTypeCustomSchema,
    ) as exc_info:
        invalid_processor = Processor(
            test_variant,
            custom_schema={
                test_variant: invalid_ruleset
            }
        )
        del invalid_processor
    assert (
        exc_info
    ), "Отлавливается ошибка, когда один из элементов карты "
    "содержит некорректную схему"

    with pytest.raises(
        err.ProcessorInitWrongTypeCustomSchema,
    ) as exc_info:
        invalid_processor = Processor(
            test_variant,
            custom_schema={
                'wrong': ruleset
            }
        )
        del invalid_processor
    assert (
        exc_info
    ), "Отлавливается ошибка, когда один из элементов карты "
    "содержит некорректный ключ-вариант"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_engine_inst_cache(
    engine_types,
    variants
):
    engine_t = engine_types[0]
    e = engine_t(variants[0])
    another_e_same_variant = engine_t(variants[0])
    third_e_other_variant = engine_t(variants[1])
    assert e is another_e_same_variant
    assert another_e_same_variant is not third_e_other_variant
    assert e is not third_e_other_variant



import json
from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.schema.pipe_methods import (
    ext_tracer_out_point,
    pipe_head,
    pipe_sep_string,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
    ext_api_kind,
)


ruleset = (
    Rule(
        description="Сохранить тело сообщения",
        destination=Destination(DestinationSection.REUSE, 'message'),
        pipeline=(
            pipe_head,
            lambda data: json.loads(data.decode()),
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "body",
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Сохранить заголовки сообщения",
        destination=Destination(DestinationSection.REUSE, 'message_properties'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "properties",
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Сохранить конверт сообщения",
        destination=Destination(DestinationSection.REUSE, 'message_envelope'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "envelope",
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Сохранить содержимое задачи",
        destination=Destination(DestinationSection.REUSE, 'task_data'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "task_data",
            ),
            Origin(
                OriginSection.POINT_ARGS, None,
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Разместить заголовки контекста трассера там, "
                    "где фабрика участков их ищет",
        destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
        pipeline=(
            pipe_head,
            lambda x: dict((k.lower(), v.decode() if hasattr(v, 'decode') else v) for k, v in (x or {}).items()),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: getattr(data['message_properties'], 'headers', {}),
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Разместить заголовки контекста трассера там, "
                    "где фабрика участков их ищет (из тела сообщения)",
        destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
        pipeline=(
            pipe_head,
            lambda x: dict((k.lower(), v) for k, v in (x or {}).items()),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['message'].get('context_headers'),
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: ("amqp", x[0].lower() or x[2], x[1]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')',)}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['message'].get('method', ''),
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['message_envelope'].exchange_name,
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['message_envelope'].routing_key,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Задать имя партнёрской точки для участка",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data.get('tracer_headers', {}).get('x-b3-servicename', ''),
            ),

        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело сообщения в аннотации",
        destination=Destination(DestinationSection.SPAN_LOGS, 'message'),
        pipeline=(
            pipe_head,
            # pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "message",
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести повторы задачи в теги",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.retry'),
        pipeline=(
            pipe_head,
            lambda task_data: f"{task_data.get('retry')}/{task_data.get('retry_max')}"
            if task_data else None,
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "task_data",
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести идентификатор задачи в теги",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.task'),
        pipeline=(
            pipe_head,
            lambda task_data: f"#{task_data.get('tid')}"
            if task_data else None,
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "task_data",
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



from functools import partial

from tracuni.define.type import Rule, Destination, DestinationSection, Origin, OriginSection, Stage
from tracuni.schema.pipe_methods import pipe_head, pipe_inject_headers, ext_out_headers


rule_amqp_out_tracer_context = Rule(
    description="Записать X-B3- заголовки в заголовки сообщения",
    destination=Destination(DestinationSection.POINT_ARGS, 'properties'),
    pipeline=(
        lambda data: (data[:3], data[3]),
        lambda data: (data[0], getattr(data[1], 'context_amqp_name', None)),
        pipe_head,
        partial(pipe_inject_headers, **{'prefix_key': 'headers'}),
    ),
    origins=(
        Origin(
            OriginSection.SPAN, None,
        ),
        Origin(
            OriginSection.POINT_ARGS, ext_out_headers,
        ),
        Origin(
            OriginSection.ADAPTER,
            lambda adapter: adapter.config.service_name,
        ),
        Origin(OriginSection.ADAPTER, "config",),
    ),
    stage=Stage.PRE,
)


rule_amqp_out_in_body_tracer_context = Rule(
    description="Записать X-B3- заголовки в заголовки сообщения",
    destination=Destination(DestinationSection.POINT_ARGS, 'message'),
    pipeline=(
        lambda data: (data[0], data[1], data[2], getattr(data[3], 'context_amqp_name', None), data[4]),
        lambda data: dict(data[4], **{data[3]: pipe_inject_headers(data[:3])}) if data[3] is not None else data[4],
    ),
    origins=(
        Origin(
            OriginSection.SPAN, None,
        ),
        Origin(
            OriginSection.POINT_ARGS, ext_out_headers,
        ),
        Origin(
            OriginSection.ADAPTER,
            lambda adapter: adapter.config.service_name,
        ),
        Origin(OriginSection.ADAPTER, "config",),
        Origin(
            OriginSection.POINT_ARGS, "message",
        ),
    ),
    stage=Stage.PRE,
)


rule_inner_out_tracer_context = Rule(
    description="Записать X-B3- заголовки в параметр точки",
    destination=Destination(DestinationSection.POINT_ARGS, 'xb3_headers'),
    pipeline=(
        pipe_inject_headers,
    ),
    origins=(
        Origin(
            OriginSection.SPAN, None,
        ),
        Origin(
            OriginSection.ADAPTER,
            lambda _: None,
        ),
        Origin(
            OriginSection.ADAPTER,
            lambda adapter: adapter.config.service_name,
        ),
    ),
    stage=Stage.PRE,
)


rule_inner_in_tracer_context = Rule(
    description="Разместить заголовки контекста трассера там, "
                "где фабрика участков их ищет",
    destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
    pipeline=(
        pipe_head,
        lambda x: dict((k.lower(), v) for k, v in (x or {}).items()),
    ),
    origins=(
        Origin(
            OriginSection.POINT_ARGS, "xb3_headers"
        ),
    ),
    stage=Stage.INIT,
)


rule_span_peer_name_from_application_name = Rule(
    description="Задать имя партнёрской точки для участка",
    destination=Destination(
        DestinationSection.SPAN_NAME,
        'remote_endpoint',
    ),
    pipeline=(
        pipe_head,
    ),
    origins=(
        Origin(
            OriginSection.ADAPTER,
            lambda adapter: adapter.config.service_name,
        ),
    ),
    stage=Stage.PRE,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Тесты модуля типов

"""

import pytest
from typing import (
    Tuple,
)


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


@pytest.fixture
def constants_module():
    """Все проверяемые константы находятся в этом модуле
    """
    from tracuni import trac_uni as constants
    return constants


@pytest.fixture
def constants_spec():
    """Имена на типы ожидаемых констант
    """
    return {
        'UNKNOWN_NAME': str,
        'TRACER_SYSTEM_NAME': str,
        'CUT_LIMIT': int,
        'CUT_LINE': str,
        'TAG_ESSENTIAL_FIELD_NAMES': Tuple,
        'TAG_ESSENTIAL_BODY_NAMES': Tuple,
        'USE_TORNADO': bool,
        'PATH_IGNORE_METHODS': Tuple,
        'PATH_IGNORE_PIECE': str,
        'CONTEXT_UID_NAME': str,
        'CONTEXT_SPAN_NAME': str,
    }


#
# #
# ####################
#
#
# ####################
# # Внутренние значения
#
#
# PATH_IGNORE_METHODS = ('decorated', 'wrapper')
# PATH_IGNORE_PIECE = r'tracer_uni/'
# CONTEXT_UID_NAME = 'span_context_uid'
# CONTEXT_SPAN_NAME = 'span_context'
#
#
# #
# ####################
@pytest.fixture
def own_attributes(constants_module):
    """Проверяем только собственные идентификаторы
    """
    return [el for el in dir(constants_module) if not el.startswith('__')]


# @pytest.mark.skip(reason="no way of constants define currently testing this")
def test_const_exist(constants_module, constants_spec):
    """Все ожидаемые константы присутствуют
     и принадлежат к требуемому типу, если применимо
    """
    # noinspection PyTypeChecker
    for c_name, c_type in constants_spec.items():
        assert \
            isinstance(getattr(constants_module, c_name, None), c_type), \
            "Should have {} with type <{}> in const".format(c_name, c_type)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_const_upper_case(own_attributes):
    """Все константы именованы в верхнем регистре
    """
    upper_cased = [el for el in own_attributes if el.isupper()]
    assert len(own_attributes) == len(upper_cased)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_no_extra_const(own_attributes, constants_spec):
    """Нет неожиданных констант
    """
    assert len(own_attributes) <= len(constants_spec)


@pytest.fixture
def tag_essential_field_names():
    from tracuni.trac_uni import TAG_ESSENTIAL_FIELD_NAMES
    return TAG_ESSENTIAL_FIELD_NAMES


def test_tag_essential_field_names_types(tag_essential_field_names):
    for el in tag_essential_field_names:
        assert (
            type(el) == str or type(el) == tuple
        )
        if type(el) == tuple:
            for sub_el in el:
                assert type(sub_el) == str


@pytest.fixture
def tag_essential_body_names():
    from tracuni.trac_uni import TAG_ESSENTIAL_BODY_NAMES
    return TAG_ESSENTIAL_BODY_NAMES


def test_tag_essential_body_names(tag_essential_body_names):
    for el in tag_essential_body_names:
        assert (
            type(el) == str
        )


@pytest.fixture
def path_ignore_methods():
    from tracuni.trac_uni import PATH_IGNORE_METHODS
    return PATH_IGNORE_METHODS


def test_path_ignore_methods(path_ignore_methods):
    for el in path_ignore_methods:
        assert (
            type(el) == str
        )



from aiohttp import web
from aiohttp.web_request import Request

from app.core.handler import BaseHandler
from app.handlers.http_handler import HttpHandler


class HttpRouter(BaseHandler):
    conn = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        transfer = HttpHandler(self)

        # Основные роуты transfer
        self.server.add_route("POST", "/q", transfer.http_in)
        self.server.add_route("POST", "/cb", transfer.callback_in)
        self.server.add_route("POST", "/err", transfer.gen_error)

        self.server.set_error_handler(self.error_handler)

    async def error_handler(self, context_span, request: Request, error: Exception) -> web.Response:
        self.app.log_err(error)
        if isinstance(error, web.HTTPException):
            return error
        try:
            error_code = int(str(error))
        except:
            error_code = 1
        return web.json_response({"code": error_code, "message": self.config['errors'][str(error)]}, status=500)



import weakref
from uuid import uuid4

from tracuni.define.const import (
    CONTEXT_UID_NAME,
)
from tracuni.misc.select_coroutine import tornado_decorator
if tornado_decorator:
    from tracuni.misc.select_coroutine import (
        run_with_stack_context,
        StackContext,
    )


class Context(object):
    class Data(object):
        def __init__(self, data_dict=None):
            self.data_dict = data_dict or None

        def __eq__(self, other):

            return self.data_dict is not None \
                and other.data_dict is not None \
                and self.data_dict.get(CONTEXT_UID_NAME) is not None \
                and other.data_dict.get(CONTEXT_UID_NAME) is not None \
                and self.data_dict.get(CONTEXT_UID_NAME)\
                ==\
                other.data_dict.get(CONTEXT_UID_NAME)

    _data = Data(data_dict={})
    _instances_registry = []

    def __init__(self, *args, **kwargs):
        self.current_data = Context.Data(*args, **kwargs)
        Context._instances_registry.append(weakref.proxy(self.current_data))
        self.old_data = None

    def __enter__(self):
        if Context._data == self.current_data:
            return

        self.old_context_data = Context.Data(
            data_dict=Context._data.data_dict,
        )

        Context._data = self.current_data

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old_data is not None:
            Context._data = self.old_data

        remove = set()
        for index, proxy in enumerate(Context._instances_registry):
            try:
                bool(proxy)
            except ReferenceError:
                remove.add(index)

        Context._instances_registry = [
            p
            for i, p in enumerate(Context._instances_registry)
            if i not in remove
        ]

    @classmethod
    def get_data(cls):
        return cls._data.data_dict


class ContextAccessor:
    @staticmethod
    def get(key):
        data_dict = Context.get_data()
        if data_dict is None:
            data_dict = {}
        return data_dict.get(key)

    @staticmethod
    def set(key, value):
        data_dict = Context.get_data()
        data_dict.setdefault(key, value)

    # noinspection PyDeprecation
    @staticmethod
    @tornado_decorator
    def run_root_with_context(coro):

        initial_context_data = {
            CONTEXT_UID_NAME: uuid4().hex,
        }

        def create_tornado_context():
            return Context(initial_context_data)

        result = yield run_with_stack_context(
            StackContext(create_tornado_context),
            lambda: coro()
        )

        return result


context_accessor = ContextAccessor()
